(*  Title:      CyclesOfFutures.thy
    ID:         $Id: CyclesOfFutures.thy,v 1.8 2009/03/24 17:07:58 flokam Exp $
    Author:     Ludovic Henrio and Florian Kammuller
                2008

    Note:       proof of absence of cycle of futures
*)

header {* Absence of cycle of futures for ASP\_fun *}

theory CyclesOfFutures imports WellFormedness begin

datatype Reference =
    RefAct ActivityRef
  | RefFut FutureRef 

primrec toSterm :: "Reference \<Rightarrow> sterm"
where
  "toSterm (RefAct A) = ActRef A"
| "toSterm (RefFut f) = FutRef f"

lemma toSterm_isRef: "isref (toSterm r)"
  by (cases r, auto)

(* defines the knowledge relationship between AO and future:
  a fut f knows a AO A (resp Fut f') if A (resp f') it is referenced 
          by the request corresponding to the fut f
  a AO A knows  a AO A (resp Fut f') if A (resp f') it is referenced 
          by the active object of A *)
constdefs 
  fcontainsf :: "configuration \<Rightarrow> (FutureRef * FutureRef) set"
  "fcontainsf c == {(f,f'). \<exists>A\<in>dom c. \<exists>E::general_context. 
                             f \<in> dom (c.\<^sub>RA) \<and> c.\<^sub>FA<f> = E\<up>FutRef f'}"

  fcontainsa :: "configuration \<Rightarrow> (FutureRef * ActivityRef) set"
  "fcontainsa c == {(f,B). \<exists>A\<in>dom c. \<exists>E::general_context. 
                            f \<in> dom (c.\<^sub>RA) \<and> c.\<^sub>FA<f> = E\<up>ActRef B}"

  acontainsf :: "configuration \<Rightarrow> (ActivityRef * FutureRef) set"
  "acontainsf c == {(A,f'). A \<in> dom c \<and> (\<exists>E::general_context. c.\<^sub>AA = E\<up>FutRef f')}"

  acontainsa :: "configuration \<Rightarrow> (ActivityRef * ActivityRef) set"
  "acontainsa c == {(A,B). A \<in> dom c \<and> (\<exists>E::general_context. c.\<^sub>AA = E\<up>ActRef B)}"

constdefs
  knows :: "configuration \<Rightarrow> (Reference * Reference) set"
  "knows c == {(RefFut f,RefFut f')| 
               f f'. \<exists>A\<in>dom c. \<exists>E::general_context. f \<in> dom (c.\<^sub>RA) \<and> c.\<^sub>FA<f> = E\<up>FutRef f'} \<union> 
              {(RefFut f, RefAct B)| 
               f B. \<exists>A\<in>dom c. \<exists>E::general_context. f \<in> dom (c.\<^sub>RA) \<and> c.\<^sub>FA<f> = E\<up>ActRef B} \<union>
              {(RefAct A, RefFut f')| 
               A f'. A \<in> dom c \<and> (\<exists>E::general_context. c.\<^sub>AA = E\<up>FutRef f')} \<union>
              {(RefAct A, RefAct B)| 
               A B. A \<in> dom c \<and> (\<exists>E::general_context. c.\<^sub>AA = E\<up>ActRef B)}"

lemma wf_knows_act: "(RefAct A,r) \<in> knows c \<Longrightarrow> A \<in> dom c"
  by (auto simp: knows_def)

(* now the transitive closure, but remembering the followed path *)
inductive_set 
  knowsrec:: "configuration \<Rightarrow> (Reference * Reference * Reference list) set"
  for c :: "configuration"
where
   krDirect: "(r,r') \<in> (knows c) \<Longrightarrow> (r,r',Nil) \<in> (knowsrec c)"
 | krRec: "(r,r') \<in> (knows c) \<and> (r',r'',rL) \<in> (knowsrec c) \<Longrightarrow> (r,r'',r'#rL) \<in> (knowsrec c)"

lemma pre_knowsrecI[rule_format]: "\<forall>r r'. (r,r',rL) \<in> knowsrec c \<longrightarrow> (r,r') \<in> (knows c)\<^sup>+"
proof (induct rL)
  case Nil thus ?case 
    by (intro strip, force elim: knowsrec.cases)
next
  case Cons thus ?case 
    by (intro strip, elim knowsrec.cases, simp, force)
qed

lemma knowsrec_rev[rule_format]:
  "\<forall>r r' r''. (r,r',rL) \<in> knowsrec c \<and> (r',r'') \<in> knows c \<longrightarrow> (r,r'',rL@[r']) \<in> knowsrec c"
proof (induct rL)
  case Nil thus ?case 
    by (auto intro: knowsrec.intros elim: knowsrec.cases)
next
  case (Cons a rL) thus ?case
  proof (intro strip, elim conjE)
    fix r :: Reference and r' :: Reference and r'' :: Reference
    assume "(r,r',a#rL) \<in> knowsrec c" "(r',r'') \<in> knows c"
    show "(r,r'',(a#rL)@[r']) \<in> knowsrec c"
    proof (cases rule: knowsrec.cases[OF `(r,r',a#rL) \<in> knowsrec c`])
	(* krDirect *)
      case 1 thus ?thesis by simp
    next
	(* krRec *)
      case 2 with prems(1) `(r',r'') \<in> knows c` show ?thesis
	by (simp, force intro: knowsrec.intros elim: knowsrec.cases)
    qed
  qed
qed

(* FORMATTED:
   (?r, ?r', ?rL) \<in> knowsrec ?c \<and> (?r', ?r'') \<in> knows ?c \<Longrightarrow> (?r, ?r'', ?rL @ [?r']) \<in> knowsrec ?c*)

lemma knowsreccases_rev[rule_format]:
  "\<forall>r r' r''. (r,r'',rL@[r']) \<in> knowsrec c \<longrightarrow> (r,r',rL) \<in> knowsrec c \<and> (r',r'') \<in> knows c"
proof (induct rL)
  case Nil thus ?case by (auto intro: knowsrec.intros elim: knowsrec.cases)
next
  case (Cons a rL) thus ?case
  proof (intro strip)
    fix r :: Reference and r' :: Reference and r'' :: Reference
    assume "(r,r'',(a#rL)@[r']) \<in> knowsrec c"
    show "(r,r',a#rL) \<in> knowsrec c \<and> (r',r'') \<in> knows c"
    proof (cases rule: knowsrec.cases[OF prems(2)])
	(* krDirect *)
      case 1 thus ?thesis by simp
    next
	(* krRec *)
      case 2 with prems(1) show ?thesis
	by (simp, force intro: knowsrec.intros elim: knowsrec.cases)
    qed
  qed
qed

(* FORMATTED:
   (?r, ?r'', ?rL @ [?r']) \<in> knowsrec ?c \<Longrightarrow> (?r, ?r', ?rL) \<in> knowsrec ?c \<and> (?r', ?r'') \<in> knows ?c
*)

lemma knowsrecI[rule_format]: "(r,r') \<in> (knows c)\<^sup>+ = (\<exists>rL. (r,r',rL) \<in> knowsrec c)"
proof
  assume "(r,r') \<in> (knows c)\<^sup>+" thus "\<exists>rL. (r,r',rL) \<in> knowsrec c"
  proof (induct rule: trancl_induct)
    case base thus ?case by (force intro: knowsrec.intros)
  next
    case (step y z) thus ?case
    proof (elim exE)
      fix rL :: "Reference list" assume "(r,y,rL) \<in> knowsrec c"
      with `(y,z) \<in> knows c` knowsrec_rev[of r y rL c z] 
      have "(r,z,rL@[y]) \<in> knowsrec c" by simp
      thus ?thesis by blast
    qed
  qed
next
  assume "\<exists>rL. (r,r',rL) \<in> knowsrec c"
  then obtain rL where "(r,r',rL) \<in> knowsrec c" by auto
  from pre_knowsrecI[OF this] show "(r,r') \<in> (knows c)\<^sup>+" by assumption
qed

lemma wf_knowsrec_act: "(RefAct A,r) \<in> (knows c)\<^sup>+ \<Longrightarrow> A \<in> dom c"
proof (simp add: knowsrecI, elim exE knowsrec.cases, 
       auto simp: wf_knows_act)
  fix ra assume "(RefAct A,ra) \<in> knows c"
  from wf_knows_act[OF this] show "\<exists>a b. c A = Some (a,b)" by auto
qed

(* now we can define absene of cycle easily *)
constdefs
  nocycle :: "configuration \<Rightarrow> bool"
   "nocycle c == \<forall>r. (r,r) \<notin> (knows c)\<^sup>+"

lemma noRefAct2: "noRef a \<Longrightarrow> \<forall>(E::general_context) A. a \<noteq> E\<up>ActRef A"
  by (auto simp: noRef_act)

lemma noRefFut2: "noRef a \<Longrightarrow> \<forall>(E::general_context) f. a \<noteq> E\<up>FutRef f"
  by (auto simp: noRef_fut)

lemma decom_knowsrec: "(r,r') \<in> (knows c)\<^sup>+ = (\<exists>r''. (r,r'') \<in> knows c 
                                                  \<and> (r'' = r' \<or> (r'',r') \<in> (knows c)\<^sup>+))"
proof (auto)
  assume "(r,r') \<in> (knows c)\<^sup>+" 
  thus "\<exists>r''. (r,r'') \<in> knows c \<and> (r'' = r' \<or> (r'',r') \<in> (knows c)\<^sup>+)"
  proof (induct rule: trancl_induct)
    case base thus ?case by auto
  next
    case step thus ?case by (auto, blast intro: trancl.intros)
  qed
qed

(* no cycle in the init_config *)
lemma initial_nocycle: "noRef a \<Longrightarrow> nocycle (init_config a)"
  unfolding nocycle_def init_config_def
proof
  fix r :: Reference
  def c \<equiv> "[aref 0 \<mapsto> ([fref 0 \<mapsto> a], Obj empty (Object empty))]"
  assume "noRef a"
  {
    fix t :: sterm and E :: general_context
    assume "Obj empty (Object empty) = E\<up>t"
    hence "Obj empty (Object empty) = t" by (cases E, auto simp: not_sym[OF map_upd_nonempty])
  }
  note objlem = this

  {
    fix rL :: "Reference list" have "(r,r,rL) \<notin> knowsrec c"
    proof (induct rL)
      case Nil thus ?case proof (auto) (* ccontr *)
	assume "(r, r, []) \<in> knowsrec c" thus False
	proof (cases rule: knowsrec.cases)
	  case (krDirect ra ra') with `noRef a` show ?thesis
	    unfolding c_def knows_def
	  proof (auto simp: ActiveObject_def OneRequest_def RequestQueue_def 
	                    noRef_fut noRef_act)
	    fix E :: general_context
	    assume "Obj empty (Object empty) = E\<up>ActRef (aref 0)"
	    from objlem[OF this] show ?thesis by simp
	  qed
	next
	  case (krRec ra ra' ra'' rL) with `noRef a` show ?thesis	    
	    unfolding c_def knows_def
	    by (auto simp: ActiveObject_def OneRequest_def RequestQueue_def 
		           noRef_fut noRef_act)
	qed
      qed
    next
      case (Cons x rL) thus ?case proof (auto) (* ccontr *)
	assume "(r,r,x#rL) \<in> knowsrec c" thus False
	proof (cases rule: knowsrec.cases)
	  case (krDirect ra ra') with `noRef a` show ?thesis
	    unfolding c_def knows_def
	    by (auto simp: ActiveObject_def OneRequest_def RequestQueue_def 
		           noRef_fut noRef_act)
	next
	  case (krRec ra ra' ra'' rLa) with `noRef a` show ?thesis
	    unfolding c_def knows_def
	  proof (auto simp: ActiveObject_def OneRequest_def RequestQueue_def 
		            noRef_fut noRef_act)
	    fix E :: general_context and f' :: FutureRef
	    assume "Obj empty (Object empty) = E\<up>FutRef f'"
	    from objlem[OF this] show ?thesis by simp
	  next
	    fix E :: general_context and B :: ActivityRef
	    assume "Obj empty (Object empty) = E\<up>ActRef B"
	    from objlem[OF this] show ?thesis by simp
	  qed
	qed
      qed
    qed
  }
  thus "(r, r) \<notin> (knows [aref 0 \<mapsto> ([fref 0 \<mapsto> a],Obj empty (Object empty))])\<^sup>+"
    unfolding c_def by (simp add: knowsrecI)
qed

lemma subtrancl: "A \<subseteq> A' \<Longrightarrow> A\<^sup>+ \<subseteq> A'\<^sup>+"
  by (auto, erule trancl_induct, auto)

lemma subcontains_nocycle: "\<lbrakk> nocycle c; knows c' \<subseteq> knows c \<rbrakk> \<Longrightarrow> nocycle c'"
  unfolding nocycle_def by (drule subtrancl, auto)

(* first references in the already existing active objects are unchanged by red *)
lemma red_RefAct: 
  fixes c :: configuration and c' :: configuration
  assumes "c \<rightarrow>\<^sub>\<zeta> c'"
  shows "\<forall>A\<in>dom c. ((RefAct A,r) \<in> knows c) = ((RefAct A,r) \<in> knows c')"
proof -
  have "\<forall>A\<in>dom c. c.\<^sub>AA = (c'.\<^sub>AA)"
    by (cases rule: zeta.cases[OF assms],
        auto simp: RequestQueue_def ActiveObject_def OneRequest_def)
  with actdom_and_red_inc[OF assms] show ?thesis
    unfolding knows_def RequestQueue_def ActiveObject_def OneRequest_def
    by (auto, force, blast, force+)
qed

(* consequently, if there is no created AO ... *)
lemma red_RefAct2: 
  fixes c :: configuration and c' :: configuration
  assumes "c \<rightarrow>\<^sub>\<zeta> c'" "dom c = dom c'"
  shows "((RefAct A,r) \<in> knows c) = ((RefAct A,r) \<in> knows c')"
proof -
  from red_RefAct[OF assms(1)] assms(2) 
       wf_knows_act[of A r c] wf_knows_act[of A r c']
  show ?thesis by auto
qed

lemma list_set_decomp[rule_format]: "(r \<in> set rL) \<longrightarrow> (\<exists>rL1 rL2. rL = rL1@r#rL2)"
  by (induct_tac rL, auto, rule_tac x = "a#rL1" in exI, force)

(* how to decompose a knowsrec path in the "middle" *)
lemma decomp_knowsrec[rule_format]: 
  "\<forall>r1 r2. (r1,r2,rL@r'#rL') \<in> knowsrec c 
    \<longrightarrow> ((r1,r',rL) \<in> knowsrec c \<and> (r',r2,rL') \<in> knowsrec c)"
proof (induct rL')
  case Nil thus ?case 
    by (induct_tac rL, 
        (intro strip conjI, (force intro: knowsrec.intros elim: knowsrec.cases)+)+)
next
  case Cons thus ?case 
    by (induct_tac rL, 
        (intro strip conjI, (force intro: knowsrec.intros elim: knowsrec.cases)+)+)
qed

(* FORMATTED 
 lemma decomp_knowsrec:
        (?r1.0, ?r2.0, ?rL @ ?r' # ?rL') \<in> knowsrec ?c \<Longrightarrow>
        (?r1.0, ?r', ?rL') \<in> knowsrec ?c \<and> (?r', ?r2.0, ?rL) \<in> knowsrec ?c *)


(* the other way : recomposing *)
lemma recomp_knowsrec[rule_format]: 
  "\<forall>r1 r2. ((r1,r',rL) \<in> knowsrec c \<and> (r',r2,rL') \<in> knowsrec c) 
    \<longrightarrow> (r1,r2,rL@r'#rL') \<in> knowsrec c"
proof (induct rL')
  case Nil thus ?case
    by (induct_tac rL, intro strip,
        (force intro: knowsrec.intros elim: knowsrec.cases),
	(force intro: knowsrec.intros knowsrec_rev elim: knowsrec.cases knowsreccases_rev))
next
  case Cons thus ?case
    by (induct_tac rL, intro strip,
        (force intro: knowsrec.intros elim: knowsrec.cases), 
	intro strip, elim conjE,
	(force intro: knowsrec.intros knowsrec_rev elim: knowsrec.cases knowsreccases_rev))
qed

(* FORMATTED 
 lemma recomp_knowsrec:
        (?r1.0, ?r', ?rL) \<in> knowsrec ?c \<and> (?r', ?r2.0, ?rL') \<in> knowsrec ?c \<Longrightarrow>
        (?r1.0, ?r2.0, ?rL @ ?r' # ?rL') \<in> knowsrec ?c                  *)

lemma turn_list_knowsrec[rule_format]: 
  "(r,r,rL@r'#rL') \<in> knowsrec c \<Longrightarrow> (r',r',rL'@r#rL) \<in> knowsrec c"
  by (drule decomp_knowsrec, rule  recomp_knowsrec, simp)

lemma pre_nocycle_act_upd':
  assumes "(ra, rb) \<in> knows (c(A \<mapsto> (empty, a)))"
  shows "(ra,rb) \<in> knows c \<or> (ra = RefAct A)"
proof (cases ra)
  {
    fix f Aa assume "f \<in> dom (c(A \<mapsto> (Map.empty, a)).\<^sub>RAa)"
    hence "Aa \<noteq> A" by (auto simp: RequestQueue_def)
  }note rqu_A = this

  case (RefFut f) with `(ra,rb) \<in> knows (c(A \<mapsto> (empty,a)))`
  obtain Aa 
    where "Aa \<in> dom (c(A \<mapsto> (empty,a)))"
          "f \<in> dom (c(A \<mapsto> (empty,a)).\<^sub>RAa)"
          "(\<exists>fr. rb = RefFut fr 
               \<and> (\<exists>E::general_context. (c(A \<mapsto> (empty,a)).\<^sub>FAa<f>) = E\<up>FutRef fr))
          \<or>(\<exists>Gr. rb = RefAct Gr
               \<and> (\<exists>E::general_context. (c(A \<mapsto> (empty,a)).\<^sub>FAa<f>) = E\<up>ActRef Gr))"
    unfolding knows_def by blast
  with rqu_A[OF this(2)] `ra = RefFut f`
  have "(ra,rb) \<in> knows c"
    by (cases rb, auto simp: knows_def RequestQueue_def OneRequest_def)
  thus ?thesis by simp
next
  case (RefAct G) with `(ra,rb) \<in> knows (c(A \<mapsto> (empty,a)))`
  have 
    pred_G: "G \<in> dom (c(A \<mapsto> (empty,a)))"
	    "(\<exists>f. rb = RefFut f \<and> (\<exists>E::general_context. (c(A \<mapsto> (empty,a)).\<^sub>AG) = E\<up>FutRef f)) 
	    \<or>(\<exists>B. rb = RefAct B \<and> (\<exists>E::general_context. (c(A \<mapsto> (empty,a)).\<^sub>AG) = E\<up>ActRef B))"
    unfolding knows_def by auto
  {
    assume "G \<noteq> A" with pred_G `ra = RefAct G`
    have "(ra,rb) \<in> knows c"
      by (simp add: knows_def, auto simp: ActiveObject_def)
  } with `ra = RefAct G` show ?thesis by auto
qed

lemma pre_nocycle_act_upd[rule_format]: 
  "\<forall>r r'. (r,r',rL) \<in> knowsrec (c(A \<mapsto> (empty,a))) 
    \<longrightarrow> RefAct A \<notin> set rL \<union> {r}
    \<longrightarrow> (r,r',rL) \<in> knowsrec c"
proof (induct rL)
  case Nil thus ?case 
  proof (clarify, elim knowsrec.cases, simp_all del: fun_upd_apply)
    fix ra rb assume "(ra,rb,[]) \<notin> knowsrec c"
                     "(ra,rb) \<in> knows (c(A \<mapsto> (empty,a)))"
    with pre_nocycle_act_upd'[OF this(2)] show "RefAct A = ra"
      by (auto simp: knowsrec.intros)
  qed
next
  case Cons thus ?case 
  proof (clarify, elim knowsrec.cases, simp_all del: fun_upd_apply,
         elim conjE)
    fix ra rb rc rL 
    assume "(rb,rc,rL) \<in> knowsrec (c(A \<mapsto> (empty,a)))"
           "RefAct A \<noteq> rb"
           "\<forall>r r'. (r,r',rL) \<in> knowsrec (c(A \<mapsto> (empty,a))) 
	     \<longrightarrow> RefAct A \<noteq> r \<longrightarrow> (r,r',rL) \<in> knowsrec c"
    hence "(rb,rc,rL) \<in> knowsrec c" by blast
    moreover
    assume "(ra,rc,rb#rL) \<notin> knowsrec c"
	   "(ra,rb) \<in> knows (c(A \<mapsto> (empty,a)))"
    note this(1) pre_nocycle_act_upd'[OF this(2)]
    ultimately
    show "RefAct A = ra" by (auto simp: knowsrec.intros)
  qed
qed

lemma pre_nocycle_fut_upd':
  assumes "A \<in> dom c" "(ra, rb) \<in> knows (c(A \<mapsto> ((c.\<^sub>RA)(f \<mapsto> a),c.\<^sub>AA)))"
  shows "(ra,rb) \<in> knows c \<or> (ra = RefFut f)"
proof -
  def c' \<equiv> "(c(A \<mapsto> ((c.\<^sub>RA)(f \<mapsto> a),c.\<^sub>AA)))"
  with assms(2) have "(ra,rb) \<in> knows c'" unfolding c'_def by simp
  show ?thesis
  proof (cases ra)
    {
      fix B assume "B \<in> dom c'" with `A \<in> dom c`
      have "B \<in> dom c \<and> (c.\<^sub>AB) = (c'.\<^sub>AB)"
	unfolding c'_def by (auto simp: ActiveObject_def)
    }note act = this
    case (RefAct G) with `(ra, rb) \<in> knows c'`
    have "G \<in> dom c'"
	   "(\<exists>f. rb = RefFut f \<and> (\<exists>E::general_context. (c'.\<^sub>AG) = E\<up>FutRef f)) 
	   \<or>(\<exists>B. rb = RefAct B \<and> (\<exists>E::general_context. (c'.\<^sub>AG) = E\<up>ActRef B))"
      unfolding knows_def by auto
    with act[OF this(1)] `ra = RefAct G` have "(ra,rb) \<in> knows c"
      by (cases rb, auto simp: knows_def)
    thus ?thesis by simp
  next
    case (RefFut f') with `(ra,rb) \<in> knows c'`
    obtain Aa 
      where "Aa \<in> dom c'" "f' \<in> dom (c'.\<^sub>RAa)" and
   pred_f': "(\<exists>fr. rb = RefFut fr \<and> (\<exists>E::general_context. (c'.\<^sub>FAa<f'>) = E\<up>FutRef fr))
	    \<or>(\<exists>Gr. rb = RefAct Gr \<and> (\<exists>E::general_context. (c'.\<^sub>FAa<f'>) = E\<up>ActRef Gr))"
      unfolding knows_def by blast
    from this(1) `A \<in> dom c` have "Aa \<in> dom c" unfolding c'_def by auto
    moreover
    {
      assume "f' \<noteq> f" with `f' \<in> dom (c'.\<^sub>RAa)` `A \<in> dom c`
      have "f' \<in> dom (c.\<^sub>RAa) \<and> (c.\<^sub>FAa<f'>) = (c'.\<^sub>FAa<f'>)" 
	unfolding c'_def 
	by (cases "A = Aa", auto simp: RequestQueue_def OneRequest_def)
    }
    ultimately
    show ?thesis using pred_f' `ra = RefFut f'`
      by (cases rb, auto simp: knows_def, blast+)
  qed
qed

lemma pre_nocycle_fut_upd[rule_format]: 
  "\<forall>r r'. A \<in> dom c \<longrightarrow> (r,r',rL) \<in> knowsrec (c(A \<mapsto> ((c.\<^sub>RA)(f \<mapsto> a),c.\<^sub>AA)))
    \<longrightarrow> RefFut f \<notin> set rL \<union> {r} \<longrightarrow> (r,r',rL) \<in> knowsrec c"
proof (induct rL)
  case Nil thus ?case 
  proof (clarify, elim knowsrec.cases, simp_all del: fun_upd_apply)
    fix ra rb ma b 
    assume "c A = Some (ma,b)" hence "A \<in> dom c" by auto
    assume "(ra,rb,[]) \<notin> knowsrec c"
           "(ra,rb) \<in> knows (c(A \<mapsto> ((c.\<^sub>RA)(f \<mapsto> a),c.\<^sub>AA)))"
    with pre_nocycle_fut_upd'[OF `A \<in> dom c` this(2)] show "RefFut f = ra"
      by (auto simp: knowsrec.intros)
  qed
next
  case Cons thus ?case 
  proof (clarify, elim knowsrec.cases, simp_all del: fun_upd_apply,
         elim conjE)
    fix ra rb rc rL ma b 
    assume "c A = Some (ma,b)"
	   "(rb,rc,rL) \<in> knowsrec (c(A \<mapsto> ((c.\<^sub>RA)(f \<mapsto> a),c.\<^sub>AA)))"
           "RefFut f \<noteq> rb"
           "A \<in> dom c \<longrightarrow> (\<forall>r r'. (r,r',rL) \<in> knowsrec (c(A \<mapsto> ((c.\<^sub>RA)(f \<mapsto> a),c.\<^sub>AA)))
	                    \<longrightarrow> RefFut f \<noteq> r \<longrightarrow> (r,r',rL) \<in> knowsrec c)"
    hence "(rb,rc,rL) \<in> knowsrec c" by blast
    moreover
    from `c A = Some (ma,b)` have "A \<in> dom c" by auto
    assume "(ra,rc,rb#rL) \<notin> knowsrec c"
	   "(ra,rb) \<in> knows (c(A \<mapsto> ((c.\<^sub>RA)(f \<mapsto> a),c.\<^sub>AA)))"
    note this(1) pre_nocycle_fut_upd'[OF `A \<in> dom c` this(2)]
    ultimately
    show "RefFut f = ra" by (auto simp: knowsrec.intros)
  qed
qed

lemma pre_nocycle_notin_knows_plus:
  "(ra,ra) \<notin> (knows c)\<^sup>+ \<Longrightarrow> (rb,rb,rL) \<notin> knowsrec c \<or> ra \<notin> set rL \<union> {rb}"
proof (simp add: knowsrecI, intro impI conjI, 
       force, rule notI)
  fix rb rL assume "\<forall>rL. (ra,ra,rL) \<notin> knowsrec c" 
                   "(rb,rb,rL) \<in> knowsrec c" 
  moreover
  assume "ra \<in> set rL"
  from list_set_decomp[OF this]
  obtain rL1 rL2 where "rL = rL1 @ ra # rL2"
    by auto
  ultimately show False using turn_list_knowsrec[of rb rL1 ra rL2]
    by auto
qed

lemma nocycle_act_upd:
  assumes "nocycle c" "(RefAct A,RefAct A) \<notin> (knows (c(A \<mapsto> (empty,a))))\<^sup>+"
  shows "nocycle (c(A \<mapsto> (empty,a)))"
  unfolding nocycle_def
proof (simp add: knowsrecI, intro allI)
  fix r rL 
  note pre_nocycle_notin_knows_plus[OF assms(2)]
  from this[of r rL]
  show "(r,r,rL) \<notin> knowsrec (c(A \<mapsto> (Map.empty, a)))"
  proof (auto)
    assume "(r,r,rL) \<in> knowsrec (c(A \<mapsto> (empty,a)))" 
           "RefAct A \<noteq> r" "RefAct A \<notin> set rL"
    from pre_nocycle_act_upd[OF this(1)] this(2-3)
         `nocycle c`
    show False unfolding nocycle_def by (auto simp: knowsrecI)
  qed
qed

lemma nocycle_fut_upd: 
  assumes "nocycle c" "A \<in> dom c" 
          "(RefFut f,RefFut f) \<notin> (knows (c(A \<mapsto>((c.\<^sub>RA)(f \<mapsto> a),c.\<^sub>AA))))\<^sup>+"
  shows "nocycle (c(A \<mapsto> ((c.\<^sub>RA)(f \<mapsto> a),c.\<^sub>AA)))"
  unfolding nocycle_def
proof (simp add: knowsrecI, intro allI)
  fix r rL 
  note pre_nocycle_notin_knows_plus[OF assms(3)]
  from this[of r rL]
  show "(r,r,rL) \<notin> knowsrec (c(A \<mapsto> ((c.\<^sub>RA)(f \<mapsto> a),c.\<^sub>AA)))"
  proof (auto)
    assume "(r,r,rL) \<in> knowsrec (c(A \<mapsto> ((c.\<^sub>RA)(f \<mapsto> a),c.\<^sub>AA)))" 
           "RefFut f \<noteq> r" "RefFut f \<notin> set rL"
    from pre_nocycle_fut_upd[OF `A \<in> dom c` this(1)] this(2-3)
         `nocycle c`
    show False unfolding nocycle_def by (auto simp: knowsrecI)
  qed
qed

(* minimality argument on paths: if there is a path between A and B then there is one that 
   does not pass by A and B *)
lemma minimal_path[rule_format]: 
  "(r,r',rL0) \<in> knowsrec c \<longrightarrow> (\<exists>rL. (r,r',rL) \<in> knowsrec c 
                                   \<and> r \<notin> set rL \<and> r' \<notin> set rL \<and> set rL \<subseteq> set rL0)"
proof (induct rL0 rule: length_induct, auto)
  case (1 xs)
  assume "(r,r',xs) \<in> knowsrec c"
  show "\<exists>rL. (r,r',rL) \<in> knowsrec c \<and> r \<notin> set rL \<and> r' \<notin> set rL \<and> set rL \<subseteq> set xs"
  proof (cases "r \<in> set xs")
    case True from list_set_decomp[OF this]
    obtain rL1 rL2 where "xs = rL1 @ r # rL2"
      by auto
    with `(r,r',xs) \<in> knowsrec c` decomp_knowsrec[of r r' rL1 r rL2 c]
    have "(r,r',rL2) \<in> knowsrec c" by simp
    with `xs = rL1 @ r # rL2` prems(1)
    have "\<exists>rL. (r,r',rL) \<in> knowsrec c \<and> r \<notin> set rL \<and> r' \<notin> set rL \<and> set rL \<subseteq> set rL2"
      by auto
    with `xs = rL1 @ r # rL2` show ?thesis by auto
  next
    case False show ?thesis
    proof (cases "r' \<in> set xs")
      case True from list_set_decomp[OF this]
      obtain rL1 rL2 where "xs = rL1 @ r' # rL2"
	by auto
      with `(r,r',xs) \<in> knowsrec c` decomp_knowsrec[of r r' rL1 r' rL2 c]
      have "(r,r',rL1) \<in> knowsrec c" by simp
      with `xs = rL1 @ r' # rL2` prems(1)
      have "(\<exists>rL. (r,r',rL) \<in> knowsrec c \<and> r \<notin> set rL \<and> r' \<notin> set rL \<and> set rL \<subseteq> set rL1)"
	by auto
      with `xs = rL1 @ r' # rL2` show ?thesis by auto
    next
      case False with `r \<notin> set xs` `(r,r',xs) \<in> knowsrec c`
      show ?thesis by blast
    qed
  qed
qed

lemma add_act_inc: "G \<notin> dom c \<Longrightarrow> (knows c)\<^sup>+ \<subseteq> (knows (c(G \<mapsto> (empty,a))))\<^sup>+"
  by (rule subtrancl) 
     (force simp: knows_def RequestQueue_def ActiveObject_def OneRequest_def)

lemma act_upd_knows: "\<lbrakk> (r,r') \<in> knows c; G \<notin> dom c \<rbrakk> \<Longrightarrow> (r,r') \<in> knows (c(G \<mapsto> (empty,a)))"
  by (auto simp: knows_def ActiveObject_def RequestQueue_def OneRequest_def)
     (case_tac "A = G", force, force)+

lemma not_in_dom_not_in_knows: "G \<notin> dom c \<Longrightarrow> (RefAct G,r) \<notin> knows c"
  by (auto simp: knows_def RequestQueue_def OneRequest_def ActiveObject_def)

lemma wf_knowsactrec: "(RefAct A,r,rL) \<in> knowsrec c \<Longrightarrow> A \<in> dom c"
proof (elim knowsrec.cases, auto simp: wf_knows_act)
  fix r' rL assume "(RefAct A,r') \<in> knows c"
  from wf_knows_act[OF this] show "\<exists>a b. c A = Some (a,b)"
    by auto
qed

(* wf lemmas sate relate references existing in the knows relation and in the configurtions *)
lemma wf_knows_fut: "(RefFut f,r) \<in> knows c \<Longrightarrow> \<exists>A\<in>dom c. f \<in> dom (c.\<^sub>R A)"
  by (auto simp: knows_def RequestQueue_def)

lemma wf_knowsfutrec: "(RefFut f,r,rL) \<in> knowsrec c \<Longrightarrow> \<exists>A\<in>dom c. f \<in> dom (c.\<^sub>R A)"
  by (elim knowsrec.cases, auto simp: wf_knows_fut)

lemma rev_list_notNil[rule_format]: "xs \<noteq> [] \<longrightarrow> (\<exists>x ys. xs = ys @ [x])"
  by (induct xs) auto

lemma wellformed_knows_act: 
  assumes "(r,RefAct A) \<in> knows c" "wf_conf c"
  shows "A \<in> dom c"
proof (cases r)
  case (RefAct G) with `(r,RefAct A) \<in> knows c`
  obtain E :: general_context 
    where "G \<in> dom c" "(c.\<^sub>AG) = E\<up>ActRef A"
    unfolding knows_def by auto
  with `wf_conf c` show ?thesis
    unfolding wf_conf_def wf_1a_def by auto
next
  case (RefFut f) with `(r,RefAct A) \<in> knows c`
  obtain Aa and E :: general_context 
    where "Aa \<in> dom c" "f \<in> dom (c.\<^sub>RAa)" "(c.\<^sub>FAa<f>) = E\<up>ActRef A"
    unfolding knows_def by auto
  with `wf_conf c` show ?thesis
    unfolding wf_conf_def wf_1b_def by auto
qed

lemma wellformed_knowsactrec: 
  assumes "(r,RefAct A,rL) \<in> knowsrec c" "wf_conf c"
  shows "A \<in> dom c"
proof (cases "rL = []")
  case True with `(r,RefAct A,rL) \<in> knowsrec c`
                 wellformed_knows_act[OF _ `wf_conf c`]
  show ?thesis by (auto elim: knowsrec.cases)
next
  case False 
  from rev_list_notNil[OF this] `(r,RefAct A,rL) \<in> knowsrec c`
  obtain x xs where "(r,RefAct A,xs@[x]) \<in> knowsrec c"
    by auto
  from knowsreccases_rev[OF this]
       wellformed_knows_act[OF _ `wf_conf c`]
  show ?thesis by auto
qed

lemma wellformed_knows_fut: 
  assumes "(r,RefFut f) \<in> knows c" "wf_conf c"
  shows "\<exists>A\<in>dom c. f \<in> dom (c.\<^sub>RA)"
proof (cases r)
  case (RefAct G) with `(r,RefFut f) \<in> knows c`
  obtain E :: general_context 
    where "G \<in> dom c" "(c.\<^sub>AG) = E\<up>FutRef f"
    unfolding knows_def by auto
  with `wf_conf c` show ?thesis
    unfolding wf_conf_def wf_2a_def by auto
next
  case (RefFut f') with `(r,RefFut f) \<in> knows c`
  obtain Aa and E :: general_context 
    where "Aa \<in> dom c" "f' \<in> dom (c.\<^sub>RAa)" "(c.\<^sub>FAa<f'>) = E\<up>FutRef f"
    unfolding knows_def by auto
  with `wf_conf c` show ?thesis
    unfolding wf_conf_def wf_2b_def by auto
qed

lemma wellformed_knowsfutrec: 
  assumes "(r,RefFut f,rL) \<in> knowsrec c" "wf_conf c"
  shows "\<exists>A\<in>dom c. f \<in> dom (c.\<^sub>RA)"
proof (cases "rL = []")
  case True with `(r,RefFut f,rL) \<in> knowsrec c`
                 wellformed_knows_fut[OF _ `wf_conf c`]
  show ?thesis by (auto elim: knowsrec.cases)
next
  case False 
  from rev_list_notNil[OF this] `(r,RefFut f,rL) \<in> knowsrec c`
  obtain x xs where "(r,RefFut f,xs@[x]) \<in> knowsrec c"
    by auto
  from knowsreccases_rev[OF this]
       wellformed_knows_fut[OF _ `wf_conf c`]
  show ?thesis by auto
qed

    (* updating a future reference doesn't effect the (direct) knowledge 
    of references other than the updated one *)
lemma knows_fut_upd:
  assumes "(r,r') \<in> knows cb" "A \<in> dom ca" "disjunct_futures cb" and
  def_cb: "cb = ca(A \<mapsto> ((ca.\<^sub>RA)(fk \<mapsto> t),ca.\<^sub>AA))"
  shows "(r \<noteq> RefFut fk \<longrightarrow> (r,r') \<in> knows ca)
        \<and>(r = RefFut fk \<longrightarrow> (\<exists>E::general_context. t = E\<up>(toSterm r')))"
proof -
  from assms(1) show ?thesis
  proof (cases "r = RefFut fk", simp_all)
    assume "(RefFut fk,r') \<in> knows cb"
    then obtain Aa
      where "Aa \<in> dom cb" "fk \<in> dom (cb.\<^sub>RAa)"
            "\<exists>E::general_context. cb.\<^sub>FAa<fk> = E\<up>(toSterm r')"
      unfolding knows_def by (auto,blast+)
    with `A \<in> dom ca` def_cb have "A \<in> dom cb" "fk \<in> dom (cb.\<^sub>RA)"
      by (auto simp: RequestQueue_def)
    with `disjunct_futures cb` `Aa \<in> dom cb` `fk \<in> dom(cb.\<^sub>RAa)` 
    have "Aa = A" unfolding disjunct_futures_def by auto
    with `\<exists>E::general_context. cb.\<^sub>FAa<fk> = E\<up>(toSterm r')` def_cb
    show "\<exists>E::general_context. t = E\<up>(toSterm r')" by (auto simp: OneRequest_def)
  next
    from `A \<in> dom ca` def_cb 
    have "dom ca = dom cb" "\<forall>A\<in>dom cb. (ca.\<^sub>AA) = (cb.\<^sub>AA)" 
	 "\<forall>fn. fn \<noteq> fk \<longrightarrow> (\<forall>A\<in>dom cb. (ca.\<^sub>FA<fn>) = (cb.\<^sub>FA<fn>))"
	 "\<forall>fn. fn \<noteq> fk \<longrightarrow> (\<forall>A\<in>dom cb. (fn \<in> dom (ca.\<^sub>RA)) = (fn \<in> dom (cb.\<^sub>RA)))"
      by (auto simp: ActiveObject_def RequestQueue_def OneRequest_def)
    moreover
    assume "r \<noteq> RefFut fk"
    ultimately show "(r,r') \<in> knows ca" using `(r,r') \<in> knows cb`
      unfolding knows_def 
    proof (cases r,auto)
      fix t B fn and E::general_context 
      assume "(cb.\<^sub>AB) = E\<up>FutRef fn" "cb B = Some t"
      with `\<forall>A\<in>dom cb. (ca.\<^sub>AA) = (cb.\<^sub>AA)`
      show "\<exists>E'::general_context. (ca.\<^sub>AB) = E'\<up>FutRef fn"
	by (rule_tac x = E in exI, auto, force)
    qed
  qed
qed

lemma knows_fut_upd_ctxt:
  fixes E :: general_context
  assumes "(RefFut fk,r) \<in> knows cb" "A \<in> dom ca" "fk \<in> dom (ca.\<^sub>RA)" 
          "(ca.\<^sub>FA<fk>) = E\<up>t" "disjunct_futures cb" and
  def_cb: "cb = ca(A \<mapsto> ((ca.\<^sub>RA)(fk \<mapsto> E\<up>(toSterm r')),ca.\<^sub>AA))"
  shows "(RefFut fk,r) \<in> knows ca \<or> r = r'"
proof auto
  assume "r \<noteq> r'" with diff_futactrefs diff_actfutrefs diff_actrefs diff_futrefs
  have "\<forall>E'::general_context. toSterm r' \<noteq> E'\<up>toSterm r"
    by (cases r, (cases r', auto)+)
  moreover
  from knows_fut_upd[OF `(RefFut fk,r) \<in> knows cb` `A \<in> dom ca` 
	                `disjunct_futures cb` def_cb]
  obtain E'::general_context where "E\<up>(toSterm r') = E'\<up>(toSterm r)"
    by auto
  note context_inc_Ref[OF toSterm_isRef[of r] this]
  ultimately
  have "\<exists>E'::general_context. (ca.\<^sub>FA<fk>) = E'\<up>(toSterm r)"
    using `ca.\<^sub>FA<fk> = E\<up>t` by simp
  with `A \<in> dom ca` `fk \<in> dom (ca.\<^sub>RA)`
  show "(RefFut fk,r) \<in> knows ca" unfolding knows_def by (cases r,auto)
qed

lemma fut_upd_not_in_path:
  assumes "(RefFut fk,r) \<in> (knows ca)\<^sup>+" "(r,r',rL) \<in> knowsrec cb"
          "nocycle ca" "A \<in> dom ca" and
  def_cb: "cb = ca(A \<mapsto> ((ca.\<^sub>RA)(fk \<mapsto> t),ca.\<^sub>AA))"
  shows "RefFut fk \<notin> set rL \<union> {r}"
proof -
  from `(RefFut fk,r) \<in> (knows ca)\<^sup>+` `nocycle ca`
  have "r \<noteq> RefFut fk" unfolding nocycle_def by auto
  thus ?thesis
  proof auto
    assume "RefFut fk \<in> set rL" from list_set_decomp[OF this] 
    obtain rL1 rL2 where "rL = rL1@RefFut fk#rL2"
      by auto
    with `(r,r',rL) \<in> knowsrec cb`
	 decomp_knowsrec[of r r' rL1 "RefFut fk" rL2 cb]
    have "(r,RefFut fk,rL1) \<in> knowsrec cb" by simp
    from minimal_path[OF this] obtain rL1' 
      where "(r,RefFut fk,rL1') \<in> knowsrec cb" "RefFut fk \<notin> set rL1'"
      by auto
    with `r \<noteq> RefFut fk` `A \<in> dom ca` def_cb
	 pre_nocycle_fut_upd[of A ca r "RefFut fk" rL1' fk t]
    have "(r,RefFut fk,rL1') \<in> knowsrec ca" by simp
    moreover
    from `(RefFut fk,r) \<in> (knows ca)\<^sup>+` obtain rL'
      where "(RefFut fk,r,rL') \<in> knowsrec ca" by (auto simp: knowsrecI)
    ultimately have "(RefFut fk,RefFut fk) \<in> (knows ca)\<^sup>+"
      by (simp add: knowsrecI, auto intro: knowsrec.intros,
	  rule_tac x = "rL'@r#rL1'" in exI, simp add: recomp_knowsrec)
    with `nocycle ca` show False unfolding nocycle_def by simp
  qed
qed

    (* adding an unknown activity reference doesn't effect the (direct) knowledge 
       of references other than the added one *)
lemma knows_act_upd:
  assumes "(r,r') \<in> knows cb" "G \<notin> dom ca" "wf_conf ca" and
  def_cb: "cb = ca(G \<mapsto> (empty,a))"
  shows "(r \<noteq> RefAct G \<longrightarrow> (r,r') \<in> knows ca)
        \<and>(r = RefAct G \<longrightarrow> (\<exists>E::general_context. a = E\<up>(toSterm r')))"
proof -
  from assms(1) show ?thesis
  proof (cases "r = RefAct G", simp_all)
    assume "(RefAct G,r') \<in> knows cb"
    with def_cb show "\<exists>E::general_context. a = E\<up>(toSterm r')"
      unfolding knows_def by (cases r', auto simp: ActiveObject_def)
  next
    from `G \<notin> dom ca` have "\<forall>B\<in>dom ca. B \<noteq> G" by auto
    with def_cb ActiveObject_diff[of _ G ca] RequestQueue_diff[of _ G ca]
    have "\<forall>A\<in>dom ca. (ca.\<^sub>AA) = (cb.\<^sub>AA)" "dom (cb.\<^sub>RG) = {}"
	 "\<forall>fn. \<forall>A\<in>dom ca. (ca.\<^sub>FA<fn>) = (cb.\<^sub>FA<fn>)"
	 "\<forall>A\<in>dom ca. (ca.\<^sub>RA) = (cb.\<^sub>RA)"
      by (auto simp: RequestQueue_def OneRequest_def)

    case False with `(r,r') \<in> knows cb`
    show "(r,r') \<in> knows ca"
    proof (cases r, auto)
      fix B assume "(RefAct B,r') \<in> knows cb" "B \<noteq> G"
      with def_cb have "B \<in> dom ca" "\<exists>E'::general_context. (cb.\<^sub>AB) = E'\<up>(toSterm r')"
	unfolding knows_def by auto
      with `\<forall>A\<in>dom ca. (ca.\<^sub>AA) = (cb.\<^sub>AA)`
      show "(RefAct B,r') \<in> knows ca"
	unfolding knows_def by (cases r',auto)
    next
      fix fn assume "(RefFut fn,r') \<in> knows cb"
      with `dom (cb.\<^sub>RG) = {}` def_cb
      obtain B where "B \<in> dom ca" "fn \<in> dom (cb.\<^sub>RB)" 
	             "\<exists>E'::general_context. cb.\<^sub>FB<fn> = E'\<up>(toSterm r')"
	unfolding knows_def by (cases r',auto)
      with `\<forall>fn. \<forall>A\<in>dom ca. (ca.\<^sub>FA<fn>) = (cb.\<^sub>FA<fn>)`
	   `\<forall>A\<in>dom ca. (ca.\<^sub>RA) = (cb.\<^sub>RA)`
      show "(RefFut fn,r') \<in> knows ca"
	unfolding knows_def by (cases r',auto)
    qed
  qed
qed

(* request / self_request *)
lemma nocycle_request:
  fixes E E' :: general_context
  assumes "ca \<rightarrow>\<^sub>\<zeta> cc" "nocycle ca" "disjunct_futures ca" "wf_conf ca"
          "A \<in> dom ca" "fa \<in> dom (ca.\<^sub>RA)" "(ca.\<^sub>FA<fa>) = E\<up>(E'\<up>ActRef B)"
          "B \<in> dom ca" "\<forall>A\<in>dom ca. fb \<notin> dom (ca.\<^sub>RA)" and
  fb_cc_ca: "\<And>r. (\<exists>E::general_context. tb = E\<up>(toSterm r))
             \<Longrightarrow> (\<exists>E::general_context. (ca.\<^sub>FA<fa>) = E\<up>(toSterm r))
                \<or>(\<exists>E::general_context. (cb.\<^sub>AB) = E\<up>(toSterm r))" and
  def_cc: "cc = cb(B \<mapsto> ((cb.\<^sub>RB)(fb \<mapsto> tb),(cb.\<^sub>AB)))" and
  def_cb: "cb = ca(A \<mapsto> ((ca.\<^sub>RA)(fa \<mapsto> E\<up>FutRef fb),(ca.\<^sub>AA)))"
  shows "nocycle cc"
proof -
    (* idea: (1) nocycle cb = ca(A \<mapsto> ((ca.\<^sub>RA)(fa \<mapsto> E\<up>FutRef fn),ca.\<^sub>AA))
             (1a) there is no cycle with RefFut fa in cb
             (1b) using nocycle_fut_upd shows nocycle cb
             (2) nocycle cc = cb(B \<mapsto> ((cb.\<^sub>RB)(fb \<mapsto> tb,cb.\<^sub>AB)))
             (2a) any reference known by RefFut fb in cc is known by FutRef fa in ca
             (2b) hence, fb doesn't know itself, since it would be known in ca otherwise 
             (2c) using nocycle cb nocycle_fut_upd shows nocycle cc
     *)

    (* preliminaries *)
  from `A \<in> dom ca` `fa \<in> dom (ca.\<^sub>RA)` `\<forall>A\<in>dom ca. fb \<notin> dom (ca.\<^sub>RA)`
  have "RefFut fa \<noteq> RefFut fb" by auto

  from `B \<in> dom ca` def_cb have "B \<in> dom cb" by simp

    (* any reference other than RefFut fb that is known by RefFut fa in cb
    is known by it in ca as well *)
  {
    from `A \<in> dom ca` `fa \<in> dom (ca.\<^sub>RA)` def_cb
    have "dom ca = dom cb" "\<forall>A\<in>dom cb. dom (cb.\<^sub>RA) = dom (ca.\<^sub>RA)"
      by (auto simp: RequestQueue_def)
    with `disjunct_futures ca` have "disjunct_futures cb"
      unfolding disjunct_futures_def by auto
    
    fix r assume "(RefFut fa,r) \<in> knows cb" "r \<noteq> RefFut fb"
    note knows_fut_upd_ctxt[OF this(1) `A \<in> dom ca` `fa \<in> dom (ca.\<^sub>RA)` 
	                       `(ca.\<^sub>FA<fa>) = E\<up>(E'\<up>ActRef B)` `disjunct_futures cb`]
    from this[of "RefFut fb"] `r \<noteq> RefFut fb` def_cb
    have "(RefFut fa,r) \<in> knows ca" by simp
  }note fa_knows_cb_ca = this


      (* (1) nocycle cb *)
      (* (1a) RefFut fa doesn't know itself in cb *)
  {
    (* RefFut fa doesn't know itself directly *)
    {
      have "(RefFut fa,RefFut fa) \<notin> knows cb"
      proof (intro notI)
	assume "(RefFut fa,RefFut fa) \<in> knows cb"
	from fa_knows_cb_ca[OF this `RefFut fa \<noteq> RefFut fb`]
	     `nocycle ca`
	show False unfolding nocycle_def by auto
      qed
    }
    
    (* main goal -- RefFut fa doesn't know itself in cb *)
    have "(RefFut fa,RefFut fa) \<notin> (knows cb)\<^sup>+"
    proof (intro notI, simp add: knowsrecI, elim exE knowsrec.cases, 
	   auto simp: `(RefFut fa,RefFut fa) \<notin> knows cb`)
      fix r' rL assume "(RefFut fa,r') \<in> knows cb"
      with `(RefFut fa,RefFut fa) \<notin> knows cb` have "r' \<noteq> RefFut fa" by auto
      assume "(r',RefFut fa,rL) \<in> knowsrec cb"
      from minimal_path[OF this] obtain rL'
	where "(r',RefFut fa,rL') \<in> knowsrec cb" "RefFut fa \<notin> set rL'"
	by auto
      with `r' \<noteq> RefFut fa` `A \<in> dom ca` def_cb
	   pre_nocycle_fut_upd[of A ca r' "RefFut fa" rL' fa]
      have "(r',RefFut fa,rL') \<in> knowsrec ca" by auto
      with `\<forall>A\<in>dom ca. fb \<notin> dom (ca.\<^sub>RA)` `wf_conf ca` 
	   wf_knowsfutrec[of fb "RefFut fa" rL' ca]
      have "r' \<noteq> RefFut fb" by auto
      from fa_knows_cb_ca[OF `(RefFut fa,r') \<in> knows cb` this] 
	   `(r',RefFut fa,rL') \<in> knowsrec ca`
      have "(RefFut fa,RefFut fa) \<in> (knows ca)\<^sup>+"
	by (simp add: knowsrecI, auto intro: knowsrec.intros)
      with `nocycle ca` show False unfolding nocycle_def by auto
    qed  
  }

      (* (1b) cb contains no cycle *)
  with `nocycle ca` `A \<in> dom ca` def_cb nocycle_fut_upd[of ca A fa]
  have "nocycle cb" by simp
    
      (* (2) nocycle cc *)
      (* (2a) any reference known by RefFut fb in cc is known by RefFut fa in ca *)
  { 
    (* preliminaries *)
    from disjunct_futures_SR[OF `ca \<rightarrow>\<^sub>\<zeta> cc` `disjunct_futures ca`]
    have "disjunct_futures cc" by simp
    
    from def_cb have "(ca.\<^sub>AB) = (cb.\<^sub>AB)" by (auto simp: ActiveObject_def)

    (* any reference directly known by RefFut fb in cc is known by RefFut fa in ca *)
    {
      fix r assume "(RefFut fb,r) \<in> knows cc"
      from `RefFut fa \<noteq> RefFut fb` 
           knows_fut_upd[OF this `B \<in> dom cb` `disjunct_futures cc` def_cc]
           fb_cc_ca[of r]
      have "(RefFut fa,r) \<in> (knows ca)\<^sup>+"
      proof (simp, elim disjE exE)
	fix E'::general_context assume "(ca.\<^sub>FA<fa>) = E'\<up>toSterm r"
	with  `A \<in> dom ca` `fa \<in> dom (ca.\<^sub>RA)`
	have "(RefFut fa,r) \<in> knows ca" unfolding knows_def by (cases r, auto)
	thus ?thesis by simp
      next
	fix E''::general_context assume "(cb.\<^sub>AB) = E''\<up>toSterm r"
	with `B \<in> dom ca` `(ca.\<^sub>AB) = (cb.\<^sub>AB)`
	have "(RefAct B,r) \<in> knows ca" unfolding knows_def by (cases r,auto)
	moreover
	from `(ca.\<^sub>FA<fa>) = E\<up>(E'\<up>ActRef B)`
	have "\<exists>E'::general_context. (ca.\<^sub>FA<fa>) = E'\<up>ActRef B"
	  by (rule_tac x = "E\<circ>E'" in exI, simp)
	with `A \<in> dom ca` `fa \<in> dom (ca.\<^sub>RA)`
	have "(RefFut fa,RefAct B) \<in> knows ca" unfolding knows_def by auto
	ultimately
	have "(RefFut fa,r,[RefAct B]) \<in> knowsrec ca"
	  by (auto intro: knowsrec.intros)
	with knowsrecI show ?thesis by auto
      qed
    }note fb_knows_cc_ca = this

    (* main goal -- knows\<^sup>+ cc \<guillemotright> ca *)
    fix r assume "(RefFut fb,r) \<in> (knows cc)\<^sup>+"
    hence "(RefFut fa,r) \<in> (knows ca)\<^sup>+"
    proof (simp add: knowsrecI, elim exE)
      fix rL assume "(RefFut fb,r,rL) \<in> knowsrec cc"
      from minimal_path[OF this] obtain rL' 
	where "(RefFut fb,r,rL') \<in> knowsrec cc" "RefFut fb \<notin> set rL'"
	by auto
      hence "(RefFut fa,r) \<in> (knows ca)\<^sup>+"
      proof (elim knowsrec.cases, auto simp: fb_knows_cc_ca)
	fix r' rL'' assume "(RefFut fb,r') \<in> knows cc"
	                   "(r',r,rL'') \<in> knowsrec cc"
	                   "RefFut fb \<notin> set rL''" "RefFut fb \<noteq> r'"
	from `B \<in> dom cb` this(2-4) def_cc
	     pre_nocycle_fut_upd[of B cb r' r rL'' fb]
	have "(r',r,rL'') \<in> knowsrec cb" by auto
	with def_cb
	  pre_nocycle_fut_upd[OF `A \<in> dom ca` _
	    fut_upd_not_in_path[OF fb_knows_cc_ca[OF `(RefFut fb,r') \<in> knows cc`]
                                   this `nocycle ca` `A \<in> dom ca` def_cb]]
	have "(r',r) \<in> (knows ca)\<^sup>+" by (auto simp: knowsrecI)
	with fb_knows_cc_ca[OF `(RefFut fb,r') \<in> knows cc`]
	show "(RefFut fa,r) \<in> (knows ca)\<^sup>+" by simp
      qed
      thus "\<exists>rL. (RefFut fa,r,rL) \<in> knowsrec ca"
	by (simp add: knowsrecI)
    qed
  }note knows_plus_cc_ca = this

      (* (2b) RefFut fb doesn't know itself in cc *)
  have "(RefFut fb,RefFut fb) \<notin> (knows cc)\<^sup>+"
  proof (intro notI)
    assume "(RefFut fb,RefFut fb) \<in> (knows cc)\<^sup>+"
    from knows_plus_cc_ca[OF this] knowsrecI 
    obtain rL where "(RefFut fa,RefFut fb,rL) \<in> knowsrec ca" by auto
    from wellformed_knowsfutrec[OF this `wf_conf ca`] 
         `\<forall>A\<in>dom ca. fb \<notin> dom (ca.\<^sub>RA)`
    show False by simp
  qed
      
      (* (2c) cc contains no cycle *)
  with nocycle_fut_upd[OF `nocycle cb` `B \<in> dom cb`] def_cc
  show "nocycle cc" by auto
qed

(* local / reply *)
lemma nocycle_data:
  fixes E :: general_context
  assumes "ca \<rightarrow>\<^sub>\<zeta> cb" "nocycle ca" "disjunct_futures ca" "wf_conf ca"
          "A \<in> dom ca" "fa \<in> dom (ca.\<^sub>RA)" and
  def_cb: "cb = ca(A \<mapsto> ((ca.\<^sub>RA)(fa \<mapsto> t'),(ca.\<^sub>AA)))" and
  fa_cb_ca: "\<And>r. (\<exists>E::general_context. t' = E\<up>(toSterm r))
             \<Longrightarrow> (\<exists>E::general_context. ca.\<^sub>FA<fa> = E\<up>(toSterm r))
                \<or>((RefFut fa,r') \<in> knows ca \<and> (r',r) \<in> knows ca)"
  shows "nocycle cb"
proof -
  {
    from nocycle_fut_upd[OF `nocycle ca` `A \<in> dom ca`] def_cb
    have "(\<And>rL. (RefFut fa,RefFut fa,rL) \<in> knowsrec cb \<Longrightarrow> False)
          \<Longrightarrow> nocycle cb"
      by (auto simp: knowsrecI)
  }note nc_intro = this

  {
    fix r assume "(RefFut fa, r) \<in> knows cb"
    from knows_fut_upd[OF this `A \<in> dom ca`
                          disjunct_futures_SR[OF `ca \<rightarrow>\<^sub>\<zeta> cb` `disjunct_futures ca`]
                          def_cb]
         fa_cb_ca[of r] have "(RefFut fa,r) \<in> (knows ca)\<^sup>+"
    proof (simp, elim disjE conjE, simp_all)
      assume "\<exists>E::general_context. (ca.\<^sub>FA<fa>) = E\<up>(toSterm r)"
      with `A \<in> dom ca` `fa \<in> dom (ca.\<^sub>RA)` 
      have "(RefFut fa,r) \<in> knows ca"
	unfolding knows_def by (cases r, auto)
      thus ?thesis by simp
    qed
  }note knows_cb_ca = this

	(* main goal *)
  show "nocycle cb"
  proof (intro nc_intro)
    fix rL assume "(RefFut fa,RefFut fa,rL) \<in> knowsrec cb"
    from minimal_path[OF this]
    obtain rL' where "(RefFut fa,RefFut fa,rL') \<in> knowsrec cb"
                     "RefFut fa \<notin> set rL'"
      by auto
    thus False
    proof (elim knowsrec.cases, auto)
      assume "(RefFut fa,RefFut fa) \<in> knows cb"
      from knows_cb_ca[OF this] `nocycle ca`
      show False unfolding nocycle_def by auto
    next
      fix r rL'' assume "(RefFut fa, r) \<in> knows cb" 
                        "(r,RefFut fa,rL'') \<in> knowsrec cb"
                        "RefFut fa \<noteq> r" "RefFut fa \<notin> set rL''"
      from this(2-4) def_cb pre_nocycle_fut_upd[OF `A \<in> dom ca`]
      have "(r,RefFut fa,rL'') \<in> knowsrec ca" by auto
      hence "(r,RefFut fa) \<in> (knows ca)\<^sup>+" by (auto simp: knowsrecI)
      with knows_cb_ca[OF `(RefFut fa, r) \<in> knows cb`]
      have "(RefFut fa,RefFut fa) \<in> (knows ca)\<^sup>+" by simp
      with `nocycle ca` show False by (simp add: nocycle_def)
    qed
  qed
qed

(* active / update *)
lemma nocycle_active:
  fixes E :: general_context
  assumes "ca \<rightarrow>\<^sub>\<zeta> cc" "nocycle ca" "disjunct_futures ca" "wf_conf ca"
          "A \<in> dom ca" "fa \<in> dom (ca.\<^sub>RA)" "(ca.\<^sub>FA<fa>) = E\<up>t" 
          "B \<notin> dom ca" and
  def_cb: "cb = ca(A \<mapsto> ((ca.\<^sub>RA)(fa \<mapsto> E\<up>ActRef B),(ca.\<^sub>AA)))" and
  def_cc: "cc = cb(B \<mapsto> (empty,b))" and
  b_cb_ca: "\<And>r. (\<exists>E::general_context. b = E\<up>(toSterm r))
            \<Longrightarrow> (\<exists>E::general_context. (ca.\<^sub>FA<fa>) = E\<up>(toSterm r)) 
               \<or>((RefFut fa,r') \<in> knows ca \<and> (r',r) \<in> knows ca)"
  shows "nocycle cc"
proof -
    (* idea: (1) nocycle cb = ca(A \<mapsto> ((ca.\<^sub>RA)(fk \<mapsto> E\<up>ActRef G),ca.\<^sub>AA))
             (1a) there is no cycle with RefFut fk in cb
             (1b) using nocycle_fut_upd shows nocycle cb
             (2) nocycle cc = cb(G \<mapsto> (empty,a)) = c'
             (2a) any reference known by RefAct G in cc is known by RefFut fk in ca
             (2b) hence, G doesn't know itself, since G would be known in ca otherwise 
             (2c) using nocycle ce nocycle_act_upd shows nocycle cc
     *)
  def ce \<equiv> "ca(B \<mapsto> (empty,b))"
  def cf \<equiv> "ce(A \<mapsto> ((ce.\<^sub>RA)(fa \<mapsto> E\<up>ActRef B),ce.\<^sub>AA))"

    (* preliminaries *)
  from `B \<notin> dom ca` have "\<forall>A\<in>dom ca. A \<noteq> B" by auto
  with `A \<in> dom ca` def_cc def_cb fun_upd_twist[of A B ca]
       ActiveObject_diff[of _ B ca] RequestQueue_diff[of _ B ca]
  have "cc = cf" unfolding cf_def ce_def by (auto simp: OneRequest_def)

    (* (1) nocycle cb *)
    (* (1a) RefFut fa doesn't know itself in cb *)
  {
    (* preliminaries *)
    from `A \<in> dom ca` `fa \<in> dom (ca.\<^sub>RA)` def_cb 
    have "dom ca = dom cb" "\<forall>A\<in>dom cb. dom (cb.\<^sub>RA) = dom (ca.\<^sub>RA)"
      by (auto simp: RequestQueue_def)
    with `disjunct_futures ca` have "disjunct_futures cb"
      unfolding disjunct_futures_def by auto


    (* any reference other than RefAct B that is known by RefFut fa in cb
       is known by it in ca *)
    {
      fix r assume "(RefFut fa,r) \<in> knows cb"
      note knows_fut_upd_ctxt[OF this `A \<in> dom ca` `fa \<in> dom (ca.\<^sub>RA)` 
	                         `(ca.\<^sub>FA<fa>) = E\<up>t` 
                                 `disjunct_futures cb`]
      from this[of "RefAct B"] def_cb
      have "(RefFut fa,r) \<in> knows ca \<or> r = RefAct B"
	by simp
    }note knows_cb_ca = this

    (* RefFut fa doesn't know itself directly *)
    {
      have "(RefFut fa,RefFut fa) \<notin> knows cb"
      proof (intro notI)
	assume "(RefFut fa,RefFut fa) \<in> knows cb"
	from knows_cb_ca[OF this] `nocycle ca` 
	show False unfolding nocycle_def by auto
      qed
    }
	
    (* main goal *)
    have "(RefFut fa,RefFut fa) \<notin> (knows cb)\<^sup>+"
    proof (intro notI, simp add: knowsrecI, elim exE knowsrec.cases, 
	   auto simp: `(RefFut fa,RefFut fa) \<notin> knows cb`)
      fix r' rL assume "(RefFut fa,r') \<in> knows cb"
      with `(RefFut fa,RefFut fa) \<notin> knows cb` have "r' \<noteq> RefFut fa" by auto
      assume "(r',RefFut fa,rL) \<in> knowsrec cb"
      from minimal_path[OF this] obtain rL'
	where "(r',RefFut fa,rL') \<in> knowsrec cb" "RefFut fa \<notin> set rL'"
	by auto
      with `r' \<noteq> RefFut fa` `A \<in> dom ca` def_cb
	   pre_nocycle_fut_upd[of A ca r' "RefFut fa" rL' fa]
      have "(r',RefFut fa,rL') \<in> knowsrec ca" by auto
      moreover
      with `B \<notin> dom ca` `wf_conf ca` 
	   wf_knowsactrec[of B "RefFut fa" rL' ca]
      have "r' \<noteq> RefAct B" by auto
      with knows_cb_ca[OF `(RefFut fa,r') \<in> knows cb`] 
      have "(RefFut fa,r') \<in> knows ca" by simp
      ultimately have "(RefFut fa,RefFut fa) \<in> (knows ca)\<^sup>+"
	by (simp add: knowsrecI, auto intro: knowsrec.intros)
      with `nocycle ca` show False unfolding nocycle_def by auto
    qed
  }

    (* (1b) cb contains no cycle *)
  with `nocycle ca` `A \<in> dom ca` def_cb 
       nocycle_fut_upd[of ca A fa]
  have "nocycle cb" by simp

    (* (2) nocycle cc *)
    (* (2a) any reference known by RefAct B in cc (= cf) is known by RefFut fa in ca *)
  { 
    (* preliminaries *)
    from `A \<in> dom ca` have "A \<in> dom ce" unfolding ce_def by auto
    from `cc = cf` disjunct_futures_SR[OF `ca \<rightarrow>\<^sub>\<zeta> cc` `disjunct_futures ca`]
    have "disjunct_futures cf" by simp
    
    (* knows ce \<guillemotright> ca *)
    {
      fix r assume "(RefAct B,r) \<in> knows ce"
      from knows_act_upd[OF this `B \<notin> dom ca` `wf_conf ca`]
      have "\<exists>E'::general_context. b = E'\<up>(toSterm r)"
	unfolding ce_def by auto
      from b_cb_ca[OF this] have "(RefFut fa,r) \<in> (knows ca)\<^sup>+"
      proof (elim disjE conjE, simp_all)
	assume "\<exists>E::general_context. (ca.\<^sub>FA<fa>) = E\<up>(toSterm r)"
	with `A \<in> dom ca` `fa \<in> dom (ca.\<^sub>RA)` 
	have "(RefFut fa,r) \<in> knows ca"
	  unfolding knows_def by (cases r, auto)
	thus ?thesis by simp
      qed
    }note knows_ce_ca = this

    (* knows cc \<guillemotright> ca (using ce \<guillemotright> ca) *)
    {
      fix r assume "(RefAct B,r) \<in> knows cc"
      with `cc = cf` have "(RefAct B,r) \<in> knows cf" by simp
      from knows_fut_upd[OF this `A \<in> dom ce` `disjunct_futures cf`]
      have "(RefAct B,r) \<in> knows ce"  unfolding cf_def by auto
    }note knows_cc_ca = knows_ce_ca[OF this]

    (* knows\<^sup>+ cc \<guillemotright> ca *)
    fix r assume "(RefAct B,r) \<in> (knows cc)\<^sup>+"
    hence "(RefFut fa,r) \<in> (knows ca)\<^sup>+"
    proof (simp add: knowsrecI, elim exE)
      fix rL assume "(RefAct B,r,rL) \<in> knowsrec cc"
      from minimal_path[OF this] obtain rL' 
	where "(RefAct B,r,rL') \<in> knowsrec cc" "RefAct B \<notin> set rL'"
	by auto
      hence "(RefFut fa,r) \<in> (knows ca)\<^sup>+"
      proof (elim knowsrec.cases, auto)
	assume "(RefAct B,r) \<in> knows cc"
	from knows_cc_ca[OF this] show ?thesis by assumption
      next
	fix r' rL'' assume "(RefAct B,r') \<in> knows cc"
	                   "(r',r,rL'') \<in> knowsrec cc"
	                   "RefAct B \<notin> set rL''" "RefAct B \<noteq> r'"
	from this(2-4) def_cc pre_nocycle_act_upd[of r' r rL'' cb B b]
	have "(r',r,rL'') \<in> knowsrec cb" by simp
	with knows_cc_ca[OF `(RefAct B,r') \<in> knows cc`] def_cb
	     fut_upd_not_in_path[OF _ this `nocycle ca` `A \<in> dom ca`]
	have "RefFut fa \<notin> set rL'' \<union> {r'}" by auto
	from pre_nocycle_fut_upd[OF `A \<in> dom ca` _ this]
	     `(r',r,rL'') \<in> knowsrec cb` def_cb
	have "(r',r,rL'') \<in> knowsrec ca" by simp
	hence "(r',r) \<in> (knows ca)\<^sup>+" by (auto simp: knowsrecI)
	with knows_cc_ca[OF `(RefAct B,r') \<in> knows cc`]
	show ?thesis by simp
      qed
      thus "\<exists>rL. (RefFut fa,r,rL) \<in> knowsrec ca" by (simp add: knowsrecI) 
    qed
  }note knows_plus_cc_ca = this

  (* (2b) RefAct B doesn't know itself in cc *)
  have "(RefAct B,RefAct B) \<notin> (knows cc)\<^sup>+"
    proof (intro notI)
      assume "(RefAct B,RefAct B) \<in> (knows cc)\<^sup>+"
      from knows_plus_cc_ca[OF this] knowsrecI 
      obtain rL where "(RefFut fa,RefAct B,rL) \<in> knowsrec ca" by auto
      from wellformed_knowsactrec[OF this `wf_conf ca`] `B \<notin> dom ca`
      show False by simp
    qed

    (* (2c) cc contains no cycle *)
    with nocycle_act_upd[OF `nocycle cb`] def_cc
    show "nocycle cc" by auto
qed

(* the main theorem: reduction maintains absence of cycles, 
   note that wellformedness is required here *)
theorem nocycle_SR: 
  assumes "c \<rightarrow>\<^sub>\<zeta> c'" "nocycle c" "disjunct_futures c" "wf_conf c"
  shows "nocycle c'"
  using assms
proof (cases rule: zeta.cases)
  case (zeta_local A ca fk a')
    (* zeta_local applies local reduction that doesn't create new knowledge,
       thus, zeta_local doesn't create any cycle *) 
  from `c = ca` assms have ca: "ca \<rightarrow>\<^sub>\<zeta> c'" "nocycle ca" "disjunct_futures ca" "wf_conf ca"
    by simp+
  show ?thesis
  proof (intro nocycle_data[OF ca `A \<in> dom ca` `fk \<in> dom (ca.\<^sub>RA)` 
                               `c' = ca(A \<mapsto> ((ca.\<^sub>RA)(fk \<mapsto> a'),(ca.\<^sub>AA)))`],
         elim exE, intro disjI1)
    fix r and E'::general_context
    assume "a' = E'\<up>(toSterm r)" 
    with local_and_ref[of "ca.\<^sub>FA<fk>" E' "toSterm r"] `(ca.\<^sub>FA<fk>) \<rightarrow>\<^sub>\<beta> a'`
         toSterm_isRef[of r]
    show "\<exists>E'::general_context. (ca.\<^sub>FA<fk>) = E'\<up>(toSterm r)" by simp
  qed
next
  case (active G ca A fk E a)
  note def_c' = `c' = ca(A \<mapsto> ((ca.\<^sub>RA)(fk \<mapsto> E\<up>ActRef G),ca.\<^sub>AA),G \<mapsto> (empty,a))`
  def cb \<equiv> "ca(A \<mapsto> ((ca.\<^sub>RA)(fk \<mapsto> E\<up>ActRef G),ca.\<^sub>AA))"
  def cc \<equiv> "cb(G \<mapsto> (empty,a))"

  from def_c' `c = ca` assms
  have "c' = cc" "ca \<rightarrow>\<^sub>\<zeta> cc" "nocycle ca" "disjunct_futures ca" "wf_conf ca"
    unfolding cc_def cb_def by simp+
  note nc_intro =        
    nocycle_active[OF this(2-5) `A \<in> dom ca` `fk \<in> dom (ca.\<^sub>RA)` 
                      `(ca.\<^sub>FA<fk>) = E\<up>Active a` `G \<notin> dom ca`]
  from `c' = cc`
  show "nocycle c'"
  proof (simp, intro nc_intro[of cb a], simp_all add: cb_def cc_def,
         elim exE, intro disjI1)
    fix r and E'::general_context
    assume "a = E'\<up>(toSterm r)" with `(ca.\<^sub>FA<fk>) = E\<up>Active a`
    show "\<exists>E::general_context. (ca.\<^sub>FA<fk>) = E\<up>(toSterm r)"
      by (rule_tac x = "E\<circ>(cActive E')" in exI, simp)
  qed
next
  case (request ca fn A B fk E l p)
  note def_c' = `c' = ca(A \<mapsto> ((ca.\<^sub>RA)(fk \<mapsto> E\<up>FutRef fn),ca.\<^sub>AA),
                         B \<mapsto> ((ca.\<^sub>RB)(fn \<mapsto> Call (ca.\<^sub>AB) l p),ca.\<^sub>AB))`
  def cb \<equiv> "ca(A \<mapsto> ((ca.\<^sub>RA)(fk \<mapsto> E\<up>FutRef fn),ca.\<^sub>AA))"
  def cc \<equiv> "cb(B \<mapsto> ((cb.\<^sub>RB)(fn \<mapsto> Call (cb.\<^sub>AB) l p),cb.\<^sub>AB))"
  def E' \<equiv> "cCallL cHole l p"

    (* preliminaries *)
  from `A \<noteq> B` def_c' ActiveObject_diff[of B A ca] RequestQueue_diff[of B A ca]
       disjunct_futures_SR[OF `c \<rightarrow>\<^sub>\<zeta> c'` `disjunct_futures c`]
       `B \<in> dom ca`
  have "c' = cc" unfolding cc_def cb_def by auto
  with `c = ca` assms `(ca.\<^sub>FA<fk>) = E\<up> Call (ActRef B) l p` 
  have "ca \<rightarrow>\<^sub>\<zeta> cc" "nocycle ca" "disjunct_futures ca" "wf_conf ca"
       "(ca.\<^sub>FA<fk>) = E\<up>(E'\<up>ActRef B)" unfolding E'_def
    by simp+
  note nc_intro =
    nocycle_request[OF this(1-4) `A \<in> dom ca` `fk \<in> dom (ca.\<^sub>RA)`
                       this(5) `B \<in> dom ca` `\<forall>A\<in>dom ca. fn \<notin> dom (ca.\<^sub>RA)`]

  from `c' = cc` show "nocycle c'"
  proof (simp, intro nc_intro[of "Call (cb.\<^sub>AB) l p" cb], 
	 simp_all add: cb_def cc_def, fold cb_def, elim exE)
    fix r and E'::general_context
    assume "Call (cb.\<^sub>AB) l p = E'\<up>(toSterm r)"
    from ref_context_call[OF toSterm_isRef[of r] sym[OF this]]
    show "(\<exists>E::general_context. (ca.\<^sub>FA<fk>) = E\<up>(toSterm r)) 
	 \<or>(\<exists>E::general_context. (cb.\<^sub>AB) = E \<up>(toSterm r))"
    proof (elim disjE, simp, elim exE)
      fix E''::general_context assume "p = E''\<up>toSterm r"
      with `(ca.\<^sub>FA<fk>) = E\<up>Call (ActRef B) l p` show ?thesis
	by (intro disjI1, 
	    rule_tac x = "E\<circ>(cCallR (ActRef B) l E'')" in exI, simp)
    qed
  qed
next
  case (request_self ca fn A fk E l p)
  note def_c' = `c' = ca(A \<mapsto> ((ca.\<^sub>RA)(fk \<mapsto> E\<up>FutRef fn,
                                       fn \<mapsto> Call (ca.\<^sub>AA) l p),ca.\<^sub>AA))`
  def cb \<equiv> "ca(A \<mapsto> ((ca.\<^sub>RA)(fk \<mapsto> E\<up>FutRef fn),ca.\<^sub>AA))"
  def cc \<equiv> "cb(A \<mapsto> ((cb.\<^sub>RA)(fn \<mapsto> Call (cb.\<^sub>AA) l p),cb.\<^sub>AA))"
  def E' \<equiv> "cCallL cHole l p"

    (* preliminaries *)
  from def_c' `\<forall>A\<in>dom ca. fn \<notin> dom (ca.\<^sub>RA)` `A \<in> dom ca` `fk \<in> dom (ca.\<^sub>RA)`
       disjunct_futures_SR[OF `c \<rightarrow>\<^sub>\<zeta> c'` `disjunct_futures c`]
  have "c' = cc" unfolding cc_def cb_def by (auto simp: RequestQueue_def ActiveObject_def)
  with `c = ca` assms `(ca.\<^sub>FA<fk>) = E\<up> Call (ActRef A) l p` 
  have "ca \<rightarrow>\<^sub>\<zeta> cc" "nocycle ca" "disjunct_futures ca" "wf_conf ca"
       "(ca.\<^sub>FA<fk>) = E\<up>(E'\<up>ActRef A)" unfolding E'_def
    by simp+
  note nc_intro =
    nocycle_request[OF this(1-4) `A \<in> dom ca` `fk \<in> dom (ca.\<^sub>RA)`
                       this(5) `A \<in> dom ca` `\<forall>A\<in>dom ca. fn \<notin> dom (ca.\<^sub>RA)`]

  from `c' = cc` show "nocycle c'"
  proof (simp, intro nc_intro[of "Call (cb.\<^sub>AA) l p" cb], 
	 simp_all add: cb_def cc_def, fold cb_def, elim exE)
    fix r and E'::general_context
    assume "Call (cb.\<^sub>AA) l p = E'\<up>(toSterm r)"
    from ref_context_call[OF toSterm_isRef[of r] sym[OF this]]
    show "(\<exists>E::general_context. (ca.\<^sub>FA<fk>) = E\<up>(toSterm r)) 
	 \<or>(\<exists>E::general_context. (cb.\<^sub>AA) = E \<up>(toSterm r))"
    proof (elim disjE, simp, elim exE)
      fix E''::general_context assume "p = E''\<up>toSterm r"
      with `(ca.\<^sub>FA<fk>) = E\<up>Call (ActRef A) l p` show ?thesis
	by (intro disjI1, 
	    rule_tac x = "E\<circ>(cCallR (ActRef A) l E'')" in exI, simp)
    qed
  qed
next
  case (update_AO G ca A B fk E l p)
  note def_c' = `c' = ca(A \<mapsto> ((ca.\<^sub>RA)(fk \<mapsto> E\<up>ActRef G),ca.\<^sub>AA),
                         G \<mapsto> (empty,Upd (ca.\<^sub>AB) l p))`
  def cb \<equiv> "ca(A \<mapsto> ((ca.\<^sub>RA)(fk \<mapsto> E\<up>ActRef G),ca.\<^sub>AA))"
  def cc \<equiv> "cb(G \<mapsto> (empty,Upd (ca.\<^sub>AB) l p))"

  from def_c' `c = ca` assms
  have "c' = cc" "ca \<rightarrow>\<^sub>\<zeta> cc" "nocycle ca" "disjunct_futures ca" "wf_conf ca"
    unfolding cc_def cb_def by simp+
  note nc_intro =        
    nocycle_active[OF this(2-5) `A \<in> dom ca` `fk \<in> dom (ca.\<^sub>RA)` 
                      `(ca.\<^sub>FA<fk>) = E\<up>Upd (ActRef B) l p` `G \<notin> dom ca`]
  from `c' = cc`
  show "nocycle c'"
  proof (simp, intro nc_intro[of cb "Upd (ca.\<^sub>AB) l p" "RefAct B"], 
         simp_all add: cb_def cc_def, elim exE)
    fix r and E'::general_context
    assume "Upd (ca.\<^sub>AB) l p = E'\<up>(toSterm r)" 
    from ref_context_upd[OF toSterm_isRef[of r] sym[OF this]]
    show "(\<exists>E::general_context. (ca.\<^sub>FA<fk>) = E\<up>(toSterm r)) 
         \<or>((RefFut fk,RefAct B) \<in> knows ca \<and> (RefAct B,r) \<in> knows ca)"
    proof (elim disjE exE)
      fix E'::general_context assume "(ca.\<^sub>AB) = E'\<up>(toSterm r)"
      with `B \<in> dom ca` have "(RefAct B,r) \<in> knows ca"
	unfolding knows_def by (cases r, auto)
      moreover
      from `(ca.\<^sub>FA<fk>) = E\<up>Upd (ActRef B) l p` 
      have "\<exists>E::general_context. (ca.\<^sub>FA<fk>) = E\<up>ActRef B"
	by (rule_tac x = "E\<circ>(cUpdL cHole l p)" in exI, simp)
      with `A \<in> dom ca` `fk \<in> dom (ca.\<^sub>RA)` have "(RefFut fk,RefAct B) \<in> knows ca"
	unfolding knows_def by auto
      ultimately show ?thesis by simp
    next
      fix E'::general_context assume "p = E'\<up>(toSterm r)"
      with `(ca.\<^sub>FA<fk>) = E\<up>Upd (ActRef B) l p`
      have "\<exists>E::general_context. (ca.\<^sub>FA<fk>) = E\<up>(toSterm r)"
	by (rule_tac x = "E\<circ>(cUpdR (ActRef B) l E')" in exI, simp)
      thus ?thesis by simp
    qed
  qed
next
  case (reply A ca fk E fi B)
  from `c = ca` assms have ca: "ca \<rightarrow>\<^sub>\<zeta> c'" "nocycle ca" "disjunct_futures ca" "wf_conf ca"
    by simp+
  show ?thesis
  proof (intro nocycle_data[OF ca `A \<in> dom ca` `fk \<in> dom (ca.\<^sub>RA)` 
                               `c' = ca(A \<mapsto> ((ca.\<^sub>RA)(fk \<mapsto> E\<up>(ca.\<^sub>FB<fi>)),(ca.\<^sub>AA)))`],
         elim exE)
    fix r and E'::general_context
    assume "E\<up>(ca.\<^sub>FB<fi>) = E'\<up>(toSterm r)"
    from context_inc_Ref[OF toSterm_isRef[of r] this]
    show "(\<exists>E'::general_context. (ca.\<^sub>FA<fk>) = E'\<up>(toSterm r))
         \<or>((RefFut fk,RefFut fi) \<in> knows ca \<and> (RefFut fi,r) \<in> knows ca)"
    proof (elim disjE)
      assume "\<exists>E::general_context. (ca.\<^sub>FB<fi>) = E\<up>(toSterm r)"
      with `B \<in> dom ca` `fi \<in> dom (ca.\<^sub>RB)` have "(RefFut fi,r) \<in> knows ca"
	unfolding knows_def by (cases r, auto)
      moreover
      from `A \<in> dom ca` `fk \<in> dom (ca.\<^sub>RA)` `(ca.\<^sub>FA<fk>) = E\<up>FutRef fi`
      have "(RefFut fk,RefFut fi) \<in> knows ca" unfolding knows_def by auto
      ultimately
      show ?thesis by simp
    next
      assume "\<forall>b. \<exists>E'::general_context. E\<up>b = E'\<up>(toSterm r)"
      with `(ca.\<^sub>FA<fk>) = E\<up>FutRef fi` 
      have "\<exists>E'::general_context. (ca.\<^sub>FA<fk>) = E'\<up>(toSterm r)"
	by auto
      thus ?thesis by simp
    qed
  qed
qed

lemma nocycle_rtrancl[rule_format]: 
  assumes "c \<rightarrow>\<^sub>\<zeta>\<^sup>* c'"
  shows "disjunct_futures c \<longrightarrow> wf_conf c \<longrightarrow> nocycle c \<longrightarrow> nocycle c'"
proof (induct rule: rtranclp.induct[OF assms], auto)
  fix a b c assume "a \<rightarrow>\<^sub>\<zeta>\<^sup>* b" "disjunct_futures a" "wf_conf a" 
                   "b \<rightarrow>\<^sub>\<zeta> c" "nocycle b"
  from nocycle_SR[OF this(4-5)
                     disjunct_rtrancl[OF this(1-2)] 
                     wf_conf_rtrancl[OF this(1) this(3)]]
  show "nocycle c" by assumption
qed

lemma reachable_nocycle: "c \<in> reachable_configs \<Longrightarrow> nocycle c"
  unfolding reachable_configs_def
  by (simp, elim exE conjE nocycle_rtrancl, 
      auto simp: initial_config_disjunct wf_init_conf initial_nocycle)

end